#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include "tcp_listen.h"
#include "signal.h"
#include "pop3_session.h"
#include "ssl.h"
#include "daemonize.h"

static char * port = NULL;
static int become_daemon = 0;
static char * pidfile = "/var/run/akpop3d.pid";

static void sig_chld(int signo) {
  pid_t pid;
  int stat;
  while ((pid = waitpid(-1,&stat,WNOHANG)) > 0);
}

static void usage(char * argv0) {
  fprintf(stderr,"%s: usage: %s [-s] [-p <port>] [-d]\n",argv0,argv0);
  exit(EXIT_FAILURE);
}

static void write_pid(void) {
  FILE * f = fopen(pidfile,"w+");
  if (f) {
    fprintf(f,"%u",getpid());
    fclose(f);
  }
  /* ignore errors */
}

static void delete_pid(void) {
  unlink(pidfile);
}

int main(int argc, char * argv[]) {
  int listenfd, connfd;
  pid_t childpid;
  socklen_t clilen, addrlen;
  struct sockaddr * cliaddr;

  if (argc>1) {
    int i;
    for (i=1;i<argc;i++) {
      if (strcmp(argv[i],"-s")==0) {
#if HAVE_LIBSSL
        set_use_ssl(1);
        if (port==NULL)
          port = "995"; /* POP3 over SSL */
#else
        fprintf(stderr,"Error: POP3 over SSL is not supported!\n");
        exit(1);
#endif
      } else if (strcmp(argv[i],"-p")==0) {
        if (argv[i+1]) {
          port = argv[i+1];
          i++;
          continue;
        } else {
          fprintf(stderr,"Error: missing port number to -p option.\n");
          usage(argv[0]);
        }
      } else if (strcmp(argv[i],"-d")==0) {
        become_daemon = 1;
      } else {
        usage(argv[0]);
      }
    } /* for */
  }

  if (port==NULL)
    port = "110";

  if (become_daemon) {
    daemonize();
    write_pid();
    atexit(delete_pid);
  }

  openlog(argv[0], LOG_PID, LOG_MAIL);

  listenfd = Tcp_listen(NULL, port, &addrlen);
  cliaddr = malloc(addrlen);
  if (cliaddr==NULL) {
    syslog(LOG_ERR,"virtual memory exhausted. Can't allocate %d bytes: %s\n",addrlen,strerror(errno));
    perror("virtual memory exhausted");
    exit(EXIT_FAILURE);
  }

  Signal(SIGCHLD,sig_chld);

  for(;;) {
    clilen = addrlen;
    if ((connfd = accept(listenfd,cliaddr,&clilen)) < 0) {
      if (errno == EINTR) {
        continue;
      } else {
        syslog(LOG_ERR,"accept error: %s\n",strerror(errno));
        continue;
      }
    }
    if ((childpid = fork()) == 0) {
      close(listenfd);
      pop3_session(connfd);
      exit(0);
    } else if (childpid == -1) {
      syslog(LOG_ERR,"fork error. Can't service a user: %s\n",strerror(errno));
    }
    close(connfd);
  }
  /* NOTREACHED */
  return 0;
}
